@@ -14,6 +14,7 @@ import requests
 import singer
 import singer.metrics as metrics
 from requests.exceptions import HTTPError
+from requests.models import ProtocolError
 from singer import Transformer, metadata
 from singer.utils import now, strftime, strptime_to_utc
 from tap_pendo import utils as tap_pendo_utils
@@ -21,144 +22,14 @@ from tap_pendo import utils as tap_pendo_utils
 KEY_PROPERTIES = ['id']
 BASE_URL = "https://app.pendo.io"
 
-endpoints = {
-    "account": {
-        "method": "GET",
-        "endpoint": "/api/v1/account/{accountId}"
-    },
-    "accounts": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation",
-        "data": {
-            "response": {
-                "mimeType": "application/json"
-            },
-            "request": {
-                "name": "all-accounts",
-                "pipeline": [{
-                    "source": {
-                        "accounts": "null"
-                    }
-                }],
-                "requestId": "all-accounts"
-            }
-        }
-    },
-    "features": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation",
-    },
-    "guide_events": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation",
-    },
-    "feature_events": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation",
-        "data": {
-            "response": {
-                "mimeType": "application/json"
-            },
-            "request": {
-                "pipeline": [{
-                    "source": {
-                        "featureEvents": {
-                            "featureId": "{featureId}"
-                        },
-                        "timeSeries": {
-                            "period": "dayRange",
-                            "first": 1598920967000,
-                            "last": "now()"
-                        }
-                    }
-                }]
-            }
-        }
-    },
-    "guides": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation",
-    },
-    "metadata_accounts": {
-        "method": "GET",
-        "endpoint": "/api/v1/metadata/schema/account"
-    },
-    "metadata_visitors": {
-        "method": "GET",
-        "endpoint": "/api/v1/metadata/schema/visitor"
-    },
-    "events": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation",
-    },
-    "pages": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation",
-    },
-    "page_events": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation",
-    },
-    "poll_events": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation",
-    },
-    "reports": {
-        "method": "GET",
-        "endpoint": "/api/v1/report"
-    },
-    "visitor": {
-        "method": "GET",
-        "endpoint": "/api/v1/visitor/{visitorId}"
-    },
-    "visitors": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation"
-
-    },
-    "visitor_history": {
-        "method": "GET",
-        "endpoint": "/api/v1/visitor/{visitorId}/history",
-        "headers": {
-            'content-type': 'application/x-www-form-urlencoded'
-        },
-        "params": {
-            "starttime": "start_time"
-        }
-    },
-    "track_types": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation"
-    },
-    "track_events": {
-        "method": "POST",
-        "endpoint": "/api/v1/aggregation"
-    }
-}
-
 LOGGER = singer.get_logger()
 session = requests.Session()
-
+# timeout request after 300 seconds
+REQUEST_TIMEOUT = 300
 
 def get_abs_path(path):
     return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)
 
-
-def get_url(endpoint, **kwargs):
-    return BASE_URL + endpoints[endpoint]['endpoint'].format(**kwargs)
-
-
-def get_method(endpoint):
-    return endpoints[endpoint]['method']
-
-
-def get_headers(endpoint):
-    return endpoints[endpoint].get('headers', {})
-
-
-def get_params(endpoint):
-    return endpoints[endpoint].get('params', {})
-
 # Determine absolute start and end times w/ attribution_window constraint
 # abs_start/end and window_start/end must be rounded to nearest hour or day (granularity)
 # Graph API enforces max history of 28 days
@@ -177,6 +48,7 @@ def get_absolute_start_end_time(last_dttm):
 
 
 def round_times(start=None, end=None):
+    # Return start and end time with rounded time(removed hours local)
     start_rounded = None
     end_rounded = None
     # Round min_start, max_end to hours or dates
@@ -186,6 +58,7 @@ def round_times(start=None, end=None):
 
 
 def remove_hours_local(dttm):
+    # Remove hours local from provided datetime
     new_dttm = dttm.replace(hour=0, minute=0, second=0, microsecond=0)
     return new_dttm
 
@@ -206,21 +79,51 @@ def update_currently_syncing(state, stream_name):
 class Server42xRateLimitError(Exception):
     pass
 
+
+class Endpoints():
+    endpoint = ""
+    method = ""
+    headers = {}
+    params = {}
+
+    def __init__(self, endpoint, method, headers=None, params=None):
+        self.endpoint = endpoint
+        self.method = method
+        self.headers = headers
+        self.params = params
+
+    def get_url(self, **kwargs):
+        """
+        Concatenate  and format the dynamic values to the BASE_URL
+        """
+        return BASE_URL + self.endpoint.format(**kwargs)
+
+
 class Stream():
+    """
+    Base Stream class that works as a parent for child stream classes.
+    """
     name = None
     replication_method = None
     replication_key = None
     key_properties = KEY_PROPERTIES
     stream = None
-    method = "GET"
     period = None
+    # initialized the endpoint attribute which can be overriden by child streams based on
+    # the different parameters used by the stream.
+    endpoint = Endpoints("/api/v1/aggregation", "POST")
 
     def __init__(self, config=None):
         self.config = config
 
     def send_request_get_results(self, req):
-        resp = session.send(req)
+        # Set request timeout to config param `request_timeout` value.
+        # If value is 0,"0", "" or None then it will set default to default to 300.0 seconds if not passed in config.
+        config_request_timeout = self.config.get('request_timeout')
+        request_timeout = config_request_timeout and float(config_request_timeout) or REQUEST_TIMEOUT # pylint: disable=consider-using-ternary
+        resp = session.send(req, timeout=request_timeout)
 
+        # Sleep for provided time and retry if rate limit exceeded
         if 'Too Many Requests' in resp.reason:
             retry_after = 30
             LOGGER.info("Rate limit reached. Sleeping for %s seconds",
@@ -228,18 +131,24 @@ class Stream():
             time.sleep(retry_after)
             raise Server42xRateLimitError(resp.reason)
 
-        resp.raise_for_status()
+        resp.raise_for_status() # Check for requests status and raise exception in failure
 
         dec = humps.decamelize(resp.json())
         return dec
 
+    # backoff for Timeout error is already included in "requests.exceptions.RequestException"
+    # as it is the parent class of "Timeout" error
     @backoff.on_exception(backoff.expo, (requests.exceptions.RequestException, Server42xRateLimitError),
                           max_tries=5,
                           giveup=lambda e: e.response is not None and 400 <= e.
                           response.status_code < 500,
                           factor=2)
+    @backoff.on_exception(backoff.expo, (ConnectionError, ProtocolError), # backoff error
+                          max_tries=5,
+                          factor=2)
     @tap_pendo_utils.ratelimit(1, 2)
     def request(self, endpoint, params=None, **kwargs):
+        # Set requests headers, url, methods, params and extra provided arguments
         # params = params or {}
         headers = {
             'x-pendo-integration-key': self.config['x_pendo_integration_key'],
@@ -247,13 +156,13 @@ class Stream():
         }
 
         request_kwargs = {
-            'url': get_url(endpoint, **kwargs),
-            'method': get_method(endpoint),
+            'url': self.endpoint.get_url(**kwargs),
+            'method': self.endpoint.method,
             'headers': headers,
             'params': params
         }
 
-        headers = get_headers(endpoint)
+        headers = self.endpoint.headers
         if headers:
             request_kwargs['headers'].update(headers)
 
@@ -263,7 +172,7 @@ class Stream():
         if kwargs.get('json'):
             request_kwargs['json'] = kwargs.get('json')
 
-        req = requests.Request(**request_kwargs).prepare()
+        req = requests.Request(**request_kwargs).prepare() # Prepare request
 
         LOGGER.info("%s %s %s", request_kwargs['method'],
                     request_kwargs['url'], request_kwargs['params'])
@@ -271,10 +180,13 @@ class Stream():
         return self.send_request_get_results(req)
 
     def get_bookmark(self, state, stream, default, key=None):
+        # Return default value if no bookmark present in state for provided stream
         if (state is None) or ('bookmarks' not in state):
             return default
+        # Initialize bookmark if not present
         if not state.get('bookmarks').get(stream):
             state['bookmarks'][stream] = {}
+        # Look for bookmark with key for stream
         if key:
             return (state.get('bookmarks', {}).get(stream,
                                                    {}).get(key, default))
@@ -298,6 +210,7 @@ class Stream():
 
 
     def load_shared_schema_refs(self):
+        # Load and return dictionary of referenced schemas from 'schemas/shared'
         shared_schemas_path = get_abs_path('schemas/shared')
 
         shared_file_names = [
@@ -313,6 +226,7 @@ class Stream():
         return shared_schema_refs
 
     def resolve_schema_references(self, schema, key, refs):
+        # Resolve $ref of schemas with provided referenced schema
         if isinstance(schema, dict):
             for k, v in schema.items():
                 if isinstance(v, (dict, list)):
@@ -323,7 +237,7 @@ class Stream():
 
 
     def load_schema(self):
-        refs = self.load_shared_schema_refs()
+        refs = self.load_shared_schema_refs() # Load references scheamas
 
         schema_file = "schemas/{}.json".format(self.name)
         with open(get_abs_path(schema_file)) as f:
@@ -338,17 +252,30 @@ class Stream():
         schema = self.load_schema()
         mdata = metadata.new()
 
+        # Write key properties and replication method to metadata
         mdata = metadata.write(mdata, (), 'table-key-properties',
                                self.key_properties)
         mdata = metadata.write(mdata, (), 'forced-replication-method',
                                self.replication_method)
 
+        # Write replication key to metadata
         if self.replication_key:
+            # for a certain stream like "features, track_types, pages, guides"
+            # the replication key in schema is "last_updated_at" and in class variable
+            # of stream it is "lastUpdatedAt" so rather than updating the replication key
+            # value in the class variable used "humps.decamelize" for backward compatibility
+            # as for previous syncs the value in the bookmark will contain "lastUpdatedAt"
             mdata = metadata.write(mdata, (), 'valid-replication-keys',
-                                   [self.replication_key])
+                                   [humps.decamelize(self.replication_key)])
 
+        # Make inclusion automatic for the key properties and replication keys
         for field_name in schema['properties'].keys():
-            if field_name in self.key_properties or field_name == self.replication_key:
+            # for a certain stream like "features, track_types, pages, guides"
+            # the replication key in schema is "last_updated_at" and in class variable
+            # of stream it is "lastUpdatedAt" so rather than updating the replication key
+            # value in the class variable used "humps.decamelize" for backward compatibility
+            # as for previous syncs the value in the bookmark will contain "lastUpdatedAt"
+            if field_name in self.key_properties or field_name == humps.decamelize(self.replication_key):
                 mdata = metadata.write(mdata, ('properties', field_name),
                                        'inclusion', 'automatic')
             else:
@@ -356,7 +283,7 @@ class Stream():
                                        'inclusion', 'available')
 
         # For period stream adjust schema for time period
-        if self.replication_key == 'day' or self.replication_key == 'hour':
+        if self.replication_key in ('day', 'hour'):
             if hasattr(self, 'period') and self.period == 'hourRange':
                 mdata.pop(('properties', 'day'))
             elif hasattr(self, 'period') and self.period == 'dayRange':
@@ -371,6 +298,7 @@ class Stream():
         return humps.decamelize(record)
 
     def sync_substream(self, state, parent, sub_stream, parent_response):
+        # Get bookmark from state or start date for the stream
         bookmark_date = self.get_bookmark(state, sub_stream.name,
                                           self.config.get('start_date'),
                                           sub_stream.replication_key)
@@ -399,24 +327,30 @@ class Stream():
                     break
                 i += 1
 
+        # Loop over records of parent stream
         for record in parent_response:
             try:
+                # Initialize counter and transformer with specific datetime format
                 with metrics.record_counter(
                         sub_stream.name) as counter, Transformer(
                             integer_datetime_fmt=
                             "unix-milliseconds-integer-datetime-parsing"
                         ) as transformer:
-                    stream_events = sub_stream.sync(state, new_bookmark,
+                    # syncing child streams from start date or state file date
+                    stream_events = sub_stream.sync(state, bookmark_dttm,
                                                     record.get(parent.key_properties[0]))
+                    # Loop over data of sub-stream
                     for event in stream_events:
                         counter.increment()
 
+                        # Get metadata for the stream to use in transform
                         schema_dict = sub_stream.stream.schema.to_dict()
                         stream_metadata = metadata.to_map(
                             sub_stream.stream.metadata)
 
                         transformed_event = sub_stream.transform(event)
 
+                        # Transform record as per field selection in metadata
                         try:
                             transformed_record = transformer.transform(
                                 transformed_event, schema_dict,
@@ -430,10 +364,12 @@ class Stream():
                                            indent=2))
                             raise err
 
-                        event_time = strptime_to_utc(
-                            transformed_record.get(sub_stream.replication_key))
+                        # Check for replication_value from record and if value found then use it for updating bookmark
+                        replication_value = transformed_record.get(sub_stream.replication_key)
+                        if replication_value:
+                            event_time = strptime_to_utc(replication_value)
+                            new_bookmark = max(new_bookmark, event_time)
 
-                        new_bookmark = max(new_bookmark, event_time)
                         singer.write_record(sub_stream.stream.tap_stream_id,
                                             transformed_record)
 
@@ -454,11 +390,13 @@ class Stream():
     def sync(self, state, start_date=None, key_id=None):
         stream_response = self.request(self.name, json=self.get_body())['results'] or []
 
+        # Get and intialize sub-stream for the current stream
         if STREAMS.get(SUB_STREAMS.get(self.name)):
             sub_stream = STREAMS.get(SUB_STREAMS.get(self.name))(self.config)
         else:
             sub_stream = None
 
+        # Sync substream if the current stream has sub-stream and selected in the catalog
         if stream_response and sub_stream and sub_stream.is_selected():
             self.sync_substream(state, self, sub_stream, stream_response)
 
@@ -466,6 +404,7 @@ class Stream():
         return (self.stream, stream_response)
 
     def lookback_window(self):
+        # Get lookback window from config and verify value
         lookback_window = self.config.get('lookback_window') or '0'
         if not lookback_window.isdigit():
             raise TypeError("lookback_window '{}' is not numeric. Check your configuration".format(lookback_window))
@@ -473,7 +412,11 @@ class Stream():
 
 class LazyAggregationStream(Stream):
     def send_request_get_results(self, req):
-        with session.send(req, stream=True) as resp:
+        # Set request timeout to config param `request_timeout` value.
+        # If value is 0,"0", "" or None then it will set default to default to 300.0 seconds if not passed in config.
+        config_request_timeout = self.config.get('request_timeout')
+        request_timeout = config_request_timeout and float(config_request_timeout) or REQUEST_TIMEOUT # pylint: disable=consider-using-ternary
+        with session.send(req, stream=True, timeout=request_timeout) as resp:
             if 'Too Many Requests' in resp.reason:
                 retry_after = 30
                 LOGGER.info("Rate limit reached. Sleeping for %s seconds",
@@ -481,19 +424,25 @@ class LazyAggregationStream(Stream):
                 time.sleep(retry_after)
                 raise Server42xRateLimitError(resp.reason)
 
-            resp.raise_for_status()
+            resp.raise_for_status() # Check for requests status and raise exception in failure
 
+            # Return list of records instead of yielding because more than one iteration occur over data in tap flow
+            # and yield will return generator which flushes out after one iteration.
+            to_return = []
             for item in ijson.items(resp.raw, 'results.item'):
-                yield humps.decamelize(item)
+                to_return.append(humps.decamelize(item))
+            return to_return
 
     def sync(self, state, start_date=None, key_id=None):
         stream_response = self.request(self.name, json=self.get_body()) or []
 
+        # Get and intialize sub-stream for the current stream
         if STREAMS.get(SUB_STREAMS.get(self.name)):
             sub_stream = STREAMS.get(SUB_STREAMS.get(self.name))(self.config)
         else:
             sub_stream = None
 
+        # Sync substream if the current stream has sub-stream and selected in the catalog
         if stream_response and sub_stream and sub_stream.is_selected():
             self.sync_substream(state, self, sub_stream, stream_response)
 
@@ -514,6 +463,7 @@ class EventsBase(Stream):
     def sync(self, state, start_date=None, key_id=None):
         update_currently_syncing(state, self.name)
 
+        # Calculate lookback window
         lookback = start_date - timedelta(
             days=self.lookback_window())
         ts = int(lookback.timestamp()) * 1000
@@ -531,7 +481,6 @@ class Accounts(Stream):
     replication_method = "INCREMENTAL"
     replication_key = "lastupdated"
     key_properties = ["account_id"]
-    method = "POST"
 
     def get_body(self):
         return {
@@ -551,6 +500,7 @@ class Accounts(Stream):
         }
 
     def transform(self, record):
+        # Transform data of accounts into one level dictionary with following transformation
         record['lastupdated'] = record.get('metadata').get('auto').get(
             'lastupdated')
         transformed = record
@@ -637,10 +587,12 @@ class Events(LazyAggregationStream):
                                           self.replication_key)
         bookmark_dttm = strptime_to_utc(bookmark_date)
 
+        # Set lookback window
         lookback = bookmark_dttm - timedelta(
             days=self.lookback_window())
         ts = int(lookback.timestamp()) * 1000
 
+        # Get period type from config and make request for event's data
         period = self.config.get('period')
         body = self.get_body(period, ts)
         events = self.request(self.name, json=body) or []
@@ -707,15 +659,18 @@ class PollEvents(Stream):
     def sync(self, state, start_date=None, key_id=None):
         update_currently_syncing(state, self.name)
 
+        # Get bookmark from state or start date for the stream
         bookmark_date = self.get_bookmark(state, self.name,
                                           self.config.get('start_date'),
                                           self.replication_key)
         bookmark_dttm = strptime_to_utc(bookmark_date)
 
+        # Set lookback window
         lookback = bookmark_dttm - timedelta(
             days=self.lookback_window())
         ts = int(lookback.timestamp()) * 1000
 
+        # Get period type from config and make request for event's data
         period = self.config.get('period')
         body = self.get_body(period, ts)
         events = self.request(self.name, json=body).get('results') or []
@@ -778,7 +733,7 @@ class GuideEvents(EventsBase):
                                 "period": period,
                                 "first": first,
                                 "last": "now()"
-                                }
+                            }
                         }
                     },
                     {
@@ -898,6 +853,8 @@ class Reports(Stream):
     name = "reports"
     replication_method = "INCREMENTAL"
     replication_key = "lastUpdatedAt"
+    # the endpoint attribute overriden and re-initialized with different endpoint URL and method
+    endpoint = Endpoints("/api/v1/report", "GET")
 
     def sync(self, state, start_date=None, key_id=None):
         reports = self.request(self.name)
@@ -908,6 +865,8 @@ class Reports(Stream):
 class MetadataVisitor(Stream):
     name = "metadata_visitor"
     replication_method = "FULL_TABLE"
+    # the endpoint attribute overriden and re-initialized with different endpoint URL and method
+    endpoint = Endpoints("/api/v1/metadata/schema/visitor", "GET")
 
     def sync(self, state, start_date=None, key_id=None):
         reports = self.request(self.name)
@@ -921,6 +880,16 @@ class VisitorHistory(Stream):
     replication_key = "modified_ts"
     key_properties = ['visitor_id']
     DATE_WINDOW_SIZE = 1
+    headers = {
+        'content-type': 'application/x-www-form-urlencoded'
+    }
+    params = {
+        "starttime": "start_time"
+    }
+    # the endpoint attribute overriden and re-initialized with different endpoint URL, method, headers and params
+    # the visitorId parameter will be formatted in the get_url() function of the endpoints class
+    endpoint = Endpoints(
+        "/api/v1/visitor/{visitorId}/history", "GET", headers, params)
 
     def get_params(self, start_time):
         return {"starttime": start_time}
@@ -928,15 +897,14 @@ class VisitorHistory(Stream):
     def sync(self, state, start_date=None, key_id=None):
         update_currently_syncing(state, self.name)
 
-        bookmark_date = self.get_bookmark(state, self.name,
-                                          self.config.get('start_date'),
-                                          self.replication_key)
-        bookmark_dttm = strptime_to_utc(bookmark_date)
-
-        abs_start, abs_end = get_absolute_start_end_time(bookmark_dttm)
+        # using "start_date" that is passed and not using the bookmark
+        # value stored in the state file, as it will be updated after
+        # every sync of child stream for parent stream
+        abs_start, abs_end = get_absolute_start_end_time(start_date)
         lookback = abs_start - timedelta(days=self.lookback_window())
         window_next = lookback
 
+        # Get data with sliding window upto abs_end
         while window_next <= abs_end:
             ts = int(window_next.timestamp()) * 1000
             params = self.get_params(start_time=ts)
@@ -946,7 +914,7 @@ class VisitorHistory(Stream):
             for visitor in visitor_history:
                 visitor['visitorId'] = key_id
                 yield visitor
-            window_next = window_next + timedelta(days=self.DATE_WINDOW_SIZE)
+            window_next = window_next + timedelta(days=self.DATE_WINDOW_SIZE) # Update window for next call
 
     def transform(self, record):
         max_value = max(record.get('ts', 0), record.get('last_ts', 0))
@@ -959,10 +927,6 @@ class Visitors(LazyAggregationStream):
     replication_method = "INCREMENTAL"
     replication_key = "lastupdated"
     key_properties = ["visitor_id"]
-    method = "POST"
-
-    def get_endpoint(self):
-        return "/api/v1/aggregation"
 
     def get_body(self):
         include_anonymous_visitors = bool(self.config.get('include_anonymous_visitors', 'false').lower() == 'true')
@@ -988,6 +952,7 @@ class Visitors(LazyAggregationStream):
         }
 
     def transform(self, record):
+        # Transform data of accounts into one level dictionary with following transformation
         record['lastupdated'] = record.get('metadata').get('auto').get(
             'lastupdated')
         transformed = record
@@ -1003,6 +968,8 @@ class MetadataAccounts(Stream):
     name = "metadata_accounts"
     replication_method = "FULL_TABLE"
     key_properties = []
+    # the endpoint attribute overriden and re-initialized with different endpoint URL and method
+    endpoint = Endpoints("/api/v1/metadata/schema/account", "GET")
 
     def get_body(self):
         return None
@@ -1010,11 +977,13 @@ class MetadataAccounts(Stream):
     def sync(self, state, start_date=None, key_id=None):
         stream_response = self.request(self.name, json=self.get_body())
 
+        # Get and intialize sub-stream for the current stream
         if STREAMS.get(SUB_STREAMS.get(self.name)):
             sub_stream = STREAMS.get(SUB_STREAMS.get(self.name))(self.config)
         else:
             sub_stream = None
 
+        # Sync substream if the current stream has sub-stream and selected in the catalog
         if stream_response and sub_stream and sub_stream.is_selected():
             self.sync_substream(state, self, sub_stream, stream_response)
 
@@ -1024,10 +993,13 @@ class MetadataAccounts(Stream):
     def get_fields(self):
         return self.request(self.name, json=self.get_body())
 
+
 class MetadataVisitors(Stream):
     name = "metadata_visitors"
     replication_method = "FULL_TABLE"
     key_properties = []
+    # the endpoint attribute overriden and re-initialized with different endpoint URL and method
+    endpoint = Endpoints("/api/v1/metadata/schema/visitor", "GET")
 
     def get_body(self):
         return None
@@ -1035,11 +1007,13 @@ class MetadataVisitors(Stream):
     def sync(self, state, start_date=None, key_id=None):
         stream_response = self.request(self.name, json=self.get_body())
 
+        # Get and intialize sub-stream for the current stream
         if STREAMS.get(SUB_STREAMS.get(self.name)):
             sub_stream = STREAMS.get(SUB_STREAMS.get(self.name))(self.config)
         else:
             sub_stream = None
 
+        # Sync substream if the current stream has sub-stream and selected in the catalog
         if stream_response and sub_stream and sub_stream.is_selected():
             self.sync_substream(state, self, sub_stream, stream_response)
 
